/*
 *  FDC ANZXeXg
 *
 */

#include <stdio.h>
#include <stdlib.h>
#define __DOS_INLINE__
#include <sys/dos.h>


struct DMAREG {
    unsigned char	csr;
    unsigned char	cer;
    unsigned short	spare1;
    unsigned char	dcr;
    unsigned char	ocr;
    unsigned char	scr;
    unsigned char	ccr;
    unsigned short	spare2;
    unsigned short	mtc;
    unsigned char	*mar;
    unsigned long	spare3;
    unsigned char	*dar;
    unsigned short	spare4;
    unsigned short	btc;
    unsigned char	*bar;
    unsigned long	spare5;
    unsigned char	spare6;
    unsigned char	niv;
    unsigned char	spare7;
    unsigned char	eiv;
    unsigned char	spare8;
    unsigned char	mfc;
    unsigned short	spare9;
    unsigned char	spare10;
    unsigned char	cpr;
    unsigned short	spare11;
    unsigned char	spare12;
    unsigned char	dfc;
    unsigned long	spare13;
    unsigned short	spare14;
    unsigned char	spare15;
    unsigned char	bfc;
    unsigned long	spare16;
    unsigned char	spare17;
    unsigned char	gcr;
};
volatile struct DMAREG *dma;

volatile unsigned char *fdc_stat = (unsigned char *)0xe94001;
volatile unsigned char *fdc_data = (unsigned char *)0xe94003;
volatile unsigned char *fdd_sel  = (unsigned char *)0xe94007;
volatile unsigned char *int_stat = (unsigned char *)0xe9c001;

#define BUFSIZE 0x400
unsigned char diskbuf[BUFSIZE];


void fd_wait_ready();
void fd_seek();
void motor_on();
void motor_off();
unsigned int fdc_sense_int_stat();
void fdc_int_mask();
void fdc_int_umask();
void fdc_send_command();
void fdc_read_status();
void fdc_send();
unsigned int fdc_read();
void dma_setup();
void dma_start();
void dma_stop();
void wait_complete();
void clear_flag();

int
main (int argc, char* argv[])
{
    unsigned int  i,j,block,track,sector,head;
    unsigned char c;

    block = (argc < 2) ? 0 : atoi (argv[1]);
    printf("block # = %d\n",block);
    printf(" Track  = %d\n",track  = block >> 4);
    printf(" Head   = %d\n",head   = (block & 0x8) >> 3);
    printf(" Sector = %d\n",sector = (block & 0x7)+1);

    _dos_super(0);
    fdc_int_mask();
    printf("Motor ON!\n");
    motor_on();
    printf("Wait Ready!\n");
    fd_wait_ready();
    printf("SEEK!\n");
    fd_seek(track);
    dma = (struct DMAREG *)0xe84000;
    clear_flag();
    dma_setup();
    dma_start();
    printf("READ DATA!\n");
    fdc_send_command(track, head, sector);
    printf("Wait Complete!\n");
    wait_complete();

    printf("Read Status =");
    fdc_read_status();
    for (i=0; i<BUFSIZE; i+=0x10) {
	for (j=0; j<0x10; j++)
	    printf("%02X ", diskbuf[i+j]);
	for (j=0; j<0x10; j++) {
	    c = diskbuf[i+j];
	    if ((c < 0x20) || (c >= 0xe0) || ((c >= 0x80) && (c < 0xa0)))
		printf(".");
	    else    printf("%c",diskbuf[i+j]);
	}
	printf("\n");
    }
    motor_off();
    fdc_int_umask();

    return 0;
}

void
fd_wait_ready (void)
{
    do {
	fdc_send(0x04);
	fdc_send(0x00);
    } while((fdc_read() & 0x20) == 0);
}

void
fd_seek (unsigned int track)
{
    fdc_send(0x0f);
    fdc_send(0x00);
    fdc_send(track);
    fdc_sense_int_stat();
}

unsigned int
fdc_sense_int_stat (void)
{
    unsigned int    stat;

    while(!(*int_stat & 0x80))
	;
    fdc_send(0x08);
    printf("Interrupt Status = ");
    printf("%02X ",stat = fdc_read());
    printf("%02X\n",fdc_read());
    return(stat);
}

void
motor_on (void)
{
    *fdd_sel = 0x80;
}

void
motor_off (void)
{
    *fdd_sel = 0x00;
}

void
fdc_int_mask (void)
{
    *int_stat &= 0xfb;
}

void
fdc_int_umask (void)
{
    *int_stat |= 0x04;
}

void
fdc_send_command (unsigned int trk, unsigned int head, unsigned int sect)
{
    fdc_send(0x46);	/* Command	*/
    fdc_send(head <<2);	/* HD/US1/US0	*/
    fdc_send(trk);	/* Cylinder	*/
    fdc_send(head);	/* Head		*/
    fdc_send(sect);	/* Record(Sector) */
    fdc_send(0x03);	/* Num(Block Length) */
    fdc_send(0x08);	/* EOT		*/
    fdc_send(0x35);	/* GSL		*/
    fdc_send(0x00);	/* DTL(Not Used */
}

void
fdc_read_status (void)
{
    unsigned int    i;

    for (i = 0; i<0x7; i++)
	printf(" %02X",fdc_read());
    printf("\n");
}

void
fdc_send (unsigned int dat)
{
#if 0
    unsigned int    stat;
#endif

    printf("Send:-%02X\n",dat);
    while((*fdc_stat & 0xc0) != 0x80)
	;
    *fdc_data = dat;
}

unsigned int
fdc_read (void)
{
    while((*fdc_stat & 0xc0) != 0xc0)
	;
    return(*fdc_data);
}

void
dma_setup (void)
{
    dma->dcr = 0x80;
    dma->ocr = 0xb2;
    dma->scr = 0x04;
    dma->ccr = 0x00;
    dma->cpr = 0x08;
    dma->mfc = 0x05;
    dma->dfc = 0x05;

    dma->mtc = BUFSIZE;
    dma->mar = diskbuf;
    dma->dar = (unsigned char *)fdc_data;
}

void
dma_start (void)
{
    dma->ccr |= 0x80;
}

void
wait_complete (void)
{
    while(!(dma->csr & 0x90))
	;
}

void
clear_flag (void)
{
    dma->csr = 0xff;
}

/* EOF */
